(load "env.scm")
(load "parser.scm")
(load "cv.scm")

(define-struct expression ())
(define-struct (lit-exp   expression) (value))
(define-struct (var-exp   expression) (name))
(define-struct (if-exp    expression) (test then else))
(define-struct (let-exp   expression) (params args body))
(define-struct (fun-exp   expression) (params body))
(define-struct (app-exp   expression) (fun args))
(define-struct (prim-exp  expression) (prim args))
(define-struct (set!-exp  expression) (name nval))
;; make a contextual value
(define-struct (make-cv-exp expression) (ctx init))
;; display internal structure of a cv (debugging!)
(define-struct (cv-trace-exp expression) (sym))
;; scoped assignment
(define-struct (scoped-exp expression) (ctx body))
;; activate scoped assignment
(define-struct (with-scope-exp expression) (body))

;; values
(define-struct value ())
(define-struct (closure value) (params body env))
(define-struct (cv      value) (ctx vals default)) ;; vals is a list of 2-slot vectors
                                                   ;; default is a 2-slot vector


;;;; current scoping contexts 
(define *scope* '())

;; is ctx the previously-entered scoping context?
(define (reentering-scope? ctx)
  (and (not (null? *scope*))
       (eq? ctx (car *scope*))))

;; is the reference exempt of scoped assignment?
;; (either there is no scoping context, or the ref is in 
;; a protected environment
(define (not-scoped? ref cs)
  (or (null? cs) (is-in-env ref *protected*)))

;;; current protected environment
;; (bindings in a protected environment are never made contextual 
;; upon assignment)
(define *protected* (make-empty-env))


;; top-level evaluation
(define (eval-prog prog)
  (let ((expr (parse prog)))
    (strict (eval expr (make-empty-env)))))

;; evaluate expression exp in lexical environment env
(define (eval exp env)
  (cond ((lit-exp? exp) (lit-exp-value exp))
        
        ((prim-exp? exp)
         (let ((args (strict-eval-args (prim-exp-args exp) env)))
           (apply (prim-exp-prim exp) args)))
        
        ((if-exp? exp)
         (if (strict (eval (if-exp-test exp) env))
             (eval (if-exp-then exp) env)
             (eval (if-exp-else exp) env)))
         
        ((let-exp? exp)
         (let ((args   (eval-args (let-exp-args exp) env)))
           (eval-body (let-exp-body exp) (extend-env (let-exp-params exp) (by-val args) env))))
        
        ((make-cv-exp? exp)
         (make-cv (eval (make-cv-exp-ctx exp) env) '() 
                  (vector #f (eval (make-cv-exp-init exp) env))))
        
        ((cv-trace-exp? exp)
         (let ((id (cv-trace-exp-sym exp)))
           (cv-trace id env)
           ))
        
        ((with-scope-exp? exp)
         (fluid-let ((*protected* env))
           (eval-body (with-scope-exp-body exp) env)))
        
        ((scoped-exp? exp)
         (let ((ctx (eval (scoped-exp-ctx exp) env)))
           (if (reentering-scope? ctx) (eval-body (scoped-exp-body exp) env)
               (fluid-let ((*scope* (cons ctx *scope*)))
                 (eval-body (scoped-exp-body exp) env)))))
           
        ((var-exp? exp)  (deref (env-lookup (var-exp-name exp) env)))
                                   
        ((set!-exp? exp) (setval! (env-lookup (set!-exp-name exp) env) 
                                  (eval (set!-exp-nval exp) env)
                                  (reverse *scope*)))
        
        ((fun-exp? exp)  (make-closure (fun-exp-params exp) (fun-exp-body exp) env))
        
        ((app-exp? exp)   (let* ((cl   (strict (eval (app-exp-fun exp) env)))
                                 (args (eval-args (app-exp-args exp) env))
                                 (env  (extend-env (closure-params cl) (by-val args) (closure-env cl))))
                            (eval-body (closure-body cl) env)))))
        
(define (eval-body body env)
    (let loop ((exps body) (val #f))
      (if (null? exps)
          val
          (loop (cdr exps) (eval (car exps) env)))))

(define (eval-args args env)
    (map (lambda (x) (eval x env)) args))

;; strictly evaluates list of argument expressions
;; (used for primitive applications)
(define (strict-eval-args args env)
    (map (lambda (x) (strict (eval x env))) args))

;; reduces a value to an actual (non-contextual) value
(define (strict val)
  (if (cv? val) (strict (deref (cv-val val)))
      val))

;; ensures the list of arguments is passed with the by-value semantics:
;; contextual values should not imply shared state, so all contextual values
;; are copied (only the structure, not the non-contextual values)
(define (by-val l)
  (map cv-structure-copy l))

;; sets the value of r, takes into account contextual values and 
;; scoped assignments
(define (setval! r nval cs) 
  (if (not-scoped? r cs)
      (setval-ns! r nval)
      (setval-s! r nval cs)))

;; not a scoped assignment: if r points to a normal value,
;; update it, otherwise update the contextual value.
(define (setval-ns! r nval)
  (let ((val (deref r)))
    (if (cv? val)
        (cv-update! val nval)
        (setref! r nval))))

;; scoped assignment: if r points to cv, and outer-most 
;; context = context of cv, then setval! on the value to update, 
;; with one context less.
;; if it points to a normal value, make it contextual 
;; (taking into account all scoping contexts).
(define (setval-s! r nval cs)
  (let ((val (deref r)))
    (if (and (cv? val) (eq? (cv-ctx val) (car cs)))
        (setval! (cv-to-update! val) nval (cdr cs))
        (setref! r (refine val nval cs)))))
 
;; returns a ref to the value to update in a cv
;; this is either the current value of the cv, or, a new binding 
;; if there is no binding for the current ctx.
(define (cv-to-update! cv)
  (let ((key (cv-evalctx cv)))
    (let ((ref (val-at key (cv-vals cv))))
      (if ref
          ref
          (cv-new-binding! cv key (cv-get-default cv))))))

;; returns a reference to the default value of a cv
(define (cv-get-default cv)
  (vector-ref (cv-default cv) 1))

;; returns a cv that refines v0 with nval in contexts cs
;; (returns a chain of cvs, all with default v0, and last value nval)
(define (refine v0 nval cs)
  (if (null? cs) nval
      (let ((cv (make-cv (car cs) '() (vector #f v0))))
        (cv-update! cv (refine v0 nval (cdr cs)))
        cv)))


